// ------------------------------------------------------------------------
// Simple J1939 Gateway
// ------------------------------------------------------------------------
//
// IMPOPRTANT: Depending on the CAN shield used for this programming sample,
//             please make sure you set the proper CS pin in module can.cpp.
//
//  This Arduino program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.

//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
// ------------------------------------------------------------------------

#include <stdlib.h>
#include <string.h>
#include <SPI.h>

#include "mcp_can.h"
#include "can_ext.h"

// Constants
#define SA                                0x33      // Our application's source address; randomly chosen

#define GLOBALADDRESS                    	255
#define NULLADDRESS                      	254
#define ActuatorADDRESS                    19       // Extrak actuator

#define PGN_RequestMessage                0x00EA00
#define PGN_AddressClaimed               	0x00EE00
#define PGN_AddressClaimedMSB             0x00
#define PGN_AddressClaimed2ND            	0xEE
#define PGN_AddressClaimedLSB             0x00
#define PRIORITY_AddressClaim            	6
#define PRIORITY_RequestMessage           6

#define OK                                0
#define ERR                               1

char sString[256];

// ------------------------------------------------------------------------
//  Setup routine runs on power-up or reset
// ------------------------------------------------------------------------
void setup()
{
  Serial.begin(115200);

// Initialize the CAN controller

  if (canInitialize(CAN_250KBPS) == CAN_OK)  // Baud rates defined in mcp_can_dfs.h
    Serial.print("CAN Init OK.....\n\r\n\r");
  else
    Serial.print("CAN Init Failed.\n\r");
}

// ------------------------------------------------------------------------
// Main Loop - Arduino Entry Point
// ------------------------------------------------------------------------
void loop() 
{
  byte nPriority = 0;
  long lPGN = 0;
  byte nSrcAddr = 0;
  byte nDestAddr = 0;
  byte nData[8];
  int nDataLen;
  int i;
   
// Check for user input per serial monitor
//
// Usage:
//
// 's' = Initiate network scan
//
// PGN  P SA DA Data
// EA FF 6 33 FF 00EE00 (Example)
//
// '0' = Move #0    50 mm reset motion flag
// '1' = Move #1    50 mm
// '2' = Move #2   100 mm
// '3' = Move #3   200 mm

  int nCount = ReadSerialMonitorString(sString);
  if(nCount > 0)
  {
    if(nCount == 1 && sString[0] == 's')
    {
      Serial.print("Network Scan Initiated:\n\r");
      nData[0] = PGN_AddressClaimedLSB;
      nData[1] = PGN_AddressClaimed2ND;
      nData[2] = PGN_AddressClaimedMSB;
      j1939Transmit(PGN_RequestMessage, PRIORITY_RequestMessage, SA, GLOBALADDRESS, nData, 3);

    }
    else if(nCount == 1 && sString[0] == '1')
    {
       Serial.print("Move #1: 50 mm\n\r");
       lPGN = 61184;   //0x00EF00;
       nPriority = PRIORITY_RequestMessage;
       nSrcAddr = SA;
       nDestAddr = GLOBALADDRESS;  //ActuatorADDRESS;
       nDataLen = 8;
 //0x07 0xD1 0x90 0x68 0xFF 0xFF 0xFF 0xFF
       nData[0] = 0x07;
       nData[1] = 0xD1;
       nData[2] = 0x90;
       nData[3] = 0x68;
       nData[4] = 0xFF;
       nData[5] = 0xFF;
       nData[6] = 0xFF;
       nData[7] = 0xFF;
       for(i=0;i<100;i++)
       {
          j1939Transmit(lPGN, nPriority, nSrcAddr, nDestAddr, nData, nDataLen);
          Send_Command(lPGN, nPriority, nSrcAddr, nDestAddr, nData, nDataLen);
          delay(100);
      }
    }
    else if(nCount == 1 && sString[0] == '2')
    {
       Serial.print("Move #2: 100 mm\n\r");
       lPGN = 61184;   //0x00EF00;
       nPriority = PRIORITY_RequestMessage;
       nSrcAddr = SA;
       nDestAddr = ActuatorADDRESS;
       nDataLen = 8;
//0x0F 0xA1 0x90 0x68 0xFF 0xFF 0xFF 0xFF
       nData[0] = 0x0F;
       nData[1] = 0xA1;
       nData[2] = 0x90;
       nData[3] = 0x68;
       nData[4] = 0xFF;
       nData[5] = 0xFF;
       nData[6] = 0xFF;
       nData[7] = 0xFF;
       delay(1000);
       for(i=0;i<100;i++)
       {
          j1939Transmit(lPGN, nPriority, nSrcAddr, nDestAddr, nData, nDataLen);
          Send_Command(lPGN, nPriority, nSrcAddr, nDestAddr, nData, nDataLen);
          delay(100);
       }
   }
    else if(nCount == 1 && sString[0] == '3')
    {
       Serial.print("Move #3: 200 mm\n\r");
       lPGN = 61184;   //0x00EF00;
       nPriority = PRIORITY_RequestMessage;
       nSrcAddr = SA;
       nDestAddr = ActuatorADDRESS;
       nDataLen = 8;
//0x0F 0x41 0x90 0x68 0xFF 0xFF 0xFF 0xFF
       nData[0] = 0x0F;
       nData[1] = 0x41;
       nData[2] = 0x90;
       nData[3] = 0x68;
       nData[4] = 0xFF;
       nData[5] = 0xFF;
       nData[6] = 0xFF;
       nData[7] = 0xFF;
       for(i=0;i<100;i++)
       {
          j1939Transmit(lPGN, nPriority, nSrcAddr, nDestAddr, nData, nDataLen);
          Send_Command(lPGN, nPriority, nSrcAddr, nDestAddr, nData, nDataLen);
          delay(100);
       }
    }
    else if(nCount == 1 && sString[0] == '0')
    {
       Serial.print("Move #0 Reset\n\r");
       lPGN = 61184;   //0x00EF00;
       nPriority = PRIORITY_RequestMessage;
       nSrcAddr = SA;
       nDestAddr = GLOBALADDRESS;  // ActuatorADDRESS;
       nDataLen = 8;
//0x07 0xD1 0x90 0x60 0xFF 0xFF 0xFF 0xFF
//0x00 0x00 0x00 0x00 0x00 0x00 0x00 0x00
       nData[0] = 0x00;   // 0x07;
       nData[1] = 0x00;   // 0xD1;
       nData[2] = 0x00;   // 0x90;
       nData[3] = 0x00;   // 0x60;
       nData[4] = 0x00;   // 0xFF;
       nData[5] = 0x00;   // 0xFF;
       nData[6] = 0x00;   // 0xFF;
       nData[7] = 0x00;   // 0xFF;
       for(i=0;i<100;i++)
       {
          j1939Transmit(lPGN, nPriority, nSrcAddr, nDestAddr, nData, nDataLen);
          Send_Command(lPGN, nPriority, nSrcAddr, nDestAddr, nData, nDataLen);
          delay(100);
       }
    }
    else if(ParseUserCommand(sString, nCount, &lPGN, &nPriority, &nSrcAddr, &nDestAddr, nData, &nDataLen) == true)
    {

// Print some feedback for verification purposes

      sprintf(sString, "Transmitting PGN: 0x%X P: 0x%X SA: 0x%X, DA: 0x%X Data: ", (int)lPGN, nPriority, nSrcAddr, nDestAddr);
      Serial.print(sString);
      for(int nIndex = 0; nIndex < nDataLen; nIndex++)
      {
        sprintf(sString, "0x%X ", nData[nIndex]);
        Serial.print(sString);
      }
      Serial.print("\n\r");

// Transmit the J1939 message

      j1939Transmit(lPGN, nPriority, nSrcAddr, nDestAddr, nData, nDataLen);
    }
    else
    {

// Incorrect user input; print corresponding message

      Serial.print("Incorrect Data Input:");
      Serial.print(sString);
      Serial.print("\n\r\n\r");
    }
  }

// Check for received J1939 messages

  if(j1939Receive(&lPGN, &nPriority, &nSrcAddr, &nDestAddr, nData, &nDataLen) == 0)
  {
    switch(lPGN)
    {
      case PGN_AddressClaimed:
        sprintf(sString, "Addr Claimed: 0x%X %dd NAME: ", nSrcAddr, nSrcAddr);
        Serial.print(sString);
        for(int nIndex = 0; nIndex < nDataLen; nIndex++)
        {          
          sprintf(sString, "0x%X ", nData[nIndex]);
          Serial.print(sString);
        }
        Serial.print("\n\r");
        break;

      default:
        sprintf(sString, "PGN: 0x%X Src: 0x%X Dest: 0x%X ", (int)lPGN, nSrcAddr, nDestAddr);
        Serial.print(sString);
        if(nDataLen == 0 )
          Serial.print("No Data...\n\r");
        else
        {
          Serial.print("Data: ");
          for(int nIndex = 0; nIndex < nDataLen; nIndex++)
          {          
            sprintf(sString, "0x%X ", nData[nIndex]);
            Serial.print(sString);
          }
          Serial.print("\n\r");
        }
        break;
    }
  }
}

// ------------------------------------------------------------------------
// Read message from Serial Monitor
// Returns string length
// ------------------------------------------------------------------------
int ReadSerialMonitorString(char* sString)
{
  byte nCount = 0;;

  if(Serial.available() > 0)
  {
    Serial.setTimeout(100);
    nCount = Serial.readBytes(sString, 255);    
  }
  sString[nCount] = 0;     // Terminate the string
  return nCount;
}

// ------------------------------------------------------------------------
// Analyze text string as typed by user
// Returns   true = Valid Input
//          false = Invalied input
// ------------------------------------------------------------------------
bool ParseUserCommand(char* sInput, int nStrLen, long* lPGN, byte* nPriority, byte* nSrcAddr, byte* nDestAddr, byte* nData, int* nDataLen)
{
  bool bRetCode = false;
  
// Check the blank positions

  if(sInput[4] == ' ' && sInput[6] == ' ' && sInput[9] == ' ' && sInput[12] == ' ')
  {

// Determine the number of data bytes

    *nDataLen = (nStrLen - 13) / 2;
    if(*nDataLen >= 1 && *nDataLen <= 8)
    {
      if(CheckHex(sInput, nStrLen) == true)
      {
// Set the return code to valid input

        bRetCode = true;
        
// Determine the PGN

        *lPGN = (ConvertHexASCII(sInput[0]) * 0x1000)
              + (ConvertHexASCII(sInput[1]) * 0x100)
              + (ConvertHexASCII(sInput[2]) * 0x10)
              + ConvertHexASCII(sInput[3]);     
        
// Determine the priority

        *nPriority = ConvertHexASCII(sInput[5]);
        
// Determine the source address

        *nSrcAddr = (ConvertHexASCII(sInput[7]) * 0x10)
                  + ConvertHexASCII(sInput[8]);
        
// Determine the destination address

        *nDestAddr = (ConvertHexASCII(sInput[10]) * 0x10)
                   + ConvertHexASCII(sInput[11]);
        
// Determine the data bytes

        int nOffset = 13;
        for(int nIndex = 0; nIndex < *nDataLen; nIndex++)
        {
          nData[nIndex] = (ConvertHexASCII(sInput[nOffset]) * 0x10)
                        + ConvertHexASCII(sInput[nOffset + 1]);
          nOffset = nOffset + 2;
        }
      }
    }
  }

  return bRetCode;
}

// ------------------------------------------------------------------------
// Check string for hex inputs
// Returns   true = Valid Input
//          false = Invalied input
// ------------------------------------------------------------------------
bool CheckHex(char* sString, int nLen)
{
   bool bRetCode = true;

// Scan through the string
   for(int nIndex = 0; nIndex < nLen; nIndex++)
   {
      if(sString[nIndex] < '0' || sString[nIndex] > '9')
         if(sString[nIndex] < 'A' || sString[nIndex] > 'F')
            if(sString[nIndex] != ' ')
               bRetCode = false;
               break;
   }

   return bRetCode;
}

// ------------------------------------------------------------------------
// Convert single character from ASCII Hex to byte
// ------------------------------------------------------------------------
byte ConvertHexASCII(char cChar)
{
   byte nNumber = 0;

   if(cChar >= '0' && cChar <= '9')
      nNumber = (byte)cChar - 0x30;
   else if(cChar >= 'A' && cChar <= 'F')
      nNumber = (byte)cChar - 0x41 + 10;

   return nNumber;
}

// ------------------------------------------------------------------------
//
// ------------------------------------------------------------------------
void Send_Command(long lPGN, byte nPriority, byte nSrcAddr, byte nDestAddr, byte* nData, int nDataLen)
{
   sprintf(sString, "--> Submit PGN: 0x%X P: 0x%X SA: 0x%X, DA: 0x%X \n\r--> Data: ", (int)lPGN, nPriority, nSrcAddr, nDestAddr);
   Serial.print(sString);

   for(int nIndex = 0; nIndex < 8; nIndex++)
   {
     sprintf(sString, "0x%X ", nData[nIndex]);
     Serial.print(sString);
   }
   Serial.print("\n\r");
}

// ------------------------------------------------------------------------

